package main

import (
	"context"
	"flag"
	"fmt"
	"log"
	"math/rand"
	"sync"
	"time"

	"cloud.google.com/go/storage"
	"google.golang.org/api/option"
)

const (
	// Recommendation: Enable a short life cycle on your bucket to ensure test
	// data is deleted in a timely manner.
	DEFAULT_BUCKET = "your-bucket-name-here"
)

var (
	// Input parameters
	bucket            = flag.String("bucket", DEFAULT_BUCKET, "desination bucket for data.")
	timeout           = flag.Duration("timeout", 30*time.Second, "time after which to bail.")
	numFilesPerThread = flag.Int("num-files-per-thread", 1000, "Number of files each worker should process before being done.")
	numThreads        = flag.Int("num-threads", 100, "Number of threads going simultaneously.")
)

func init() {
	flag.Parse()
}

// writeToGcs just writes a bunch of bytes as a single chunk to a specified object in GCS.
func writeToGcs(ctx context.Context, storageClient *storage.Client, srcData []byte, dstBucket, dstObj string) error {
	// Set up a storage writer.
	w := storageClient.Bucket(dstBucket).Object(dstObj).NewWriter(ctx)
	w.ChunkSize = len(srcData)

	// Write to GCS.
	_, err := w.Write(srcData)
	if err != nil {
		w.CloseWithError(err)
		return err
	}

	// Close it up.
	return w.Close()
}

// Copy a file repeatedly into a GCS prefix under different names, up to max times and until ctx ends.
func multiCopyFile(ctx context.Context, storageClient *storage.Client, workerId, maxFiles int, srcData []byte, dstPrefix string) {
	for cnt := 0; cnt < maxFiles; cnt++ {
		select {
		case <-ctx.Done():
			return
		default:
			dstObj := fmt.Sprintf("%s/%d/file-%d.bin", dstPrefix, workerId, cnt)
			err := writeToGcs(ctx, storageClient, srcData, *bucket, dstObj)
			if err != nil && ctx.Err() == nil {
				log.Printf("Worker %d: Failed to copy file: %v", workerId, err)
			}
		}
	}
}

func main() {
	ctx := context.Background()

	// Set up our storage client.
	log.Println("Setting up GCS client...")
	clientOptions := make([]option.ClientOption, 0)
	storageClient, err := storage.NewClient(ctx, clientOptions...)
	if err != nil {
		log.Fatalf("Couldn't create storage client: %v\n", err)
	}

	// Generate a random 1KB byte array to write over and over to GCS.
	srcData := make([]byte, 1024)
	rand.Read(srcData)

	// Kick off numThreads goroutines that keep writing to GCS until they're done.
	dstPrefix := time.Now().Format("2006-01-02-15-04-05")
	ctx, _ = context.WithTimeout(ctx, *timeout)
	var wg sync.WaitGroup
	wg.Add(*numThreads)
	for workerId := 0; workerId < *numThreads; workerId++ {
		go func(wid int) {
			defer wg.Done()
			multiCopyFile(ctx, storageClient, wid, *numFilesPerThread, srcData, dstPrefix)
		}(workerId)
	}

	wg.Wait()
	log.Println("Done!")
}
